@@ -1,5 +1,6 @@
 /*
     Copyright (C) 2017 Kai Uwe Broulik <kde@privat.broulik.de>
+    Copyright (C) 2018 David Edmundson <davidedmundson@kde.org>
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
@@ -58,20 +59,33 @@ chrome.runtime.onMessage.addListener(function (message, sender) {
     }
 });
 
+var storage = (IS_FIREFOX ? chrome.storage.local : chrome.storage.sync);
+
+storage.get(DEFAULT_EXTENSION_SETTINGS, function (items) {
+    if (items.breezeScrollBars.enabled) {
+        loadBreezeScrollBars();
+    }
+    if (items.mpris.enabled) {
+        loadMpris();
+        if (items.mprisMediaSessions.enabled) {
+            loadMediaSessionsShim();
+        }
+    }
+});
+
 // BREEZE SCROLL BARS
 // ------------------------------------------------------------------------
 //
-if (!IS_FIREFOX) {
-    chrome.storage.sync.get(DEFAULT_EXTENSION_SETTINGS, function (items) {
-        if (items.breezeScrollBars.enabled) {
-            var linkTag = document.createElement("link");
-            linkTag.rel = "stylesheet";
-            linkTag.href =  chrome.extension.getURL("breeze-scroll-bars.css");
-            (document.head || document.documentElement).appendChild(linkTag);
-        }
-    });
+function loadBreezeScrollBars() {
+    if (!IS_FIREFOX) {
+        var linkTag = document.createElement("link");
+        linkTag.rel = "stylesheet";
+        linkTag.href =  chrome.extension.getURL("breeze-scroll-bars.css");
+        (document.head || document.documentElement).appendChild(linkTag);
+    }
 }
 
+
 // MPRIS
 // ------------------------------------------------------------------------
 //
@@ -182,6 +196,20 @@ addCallback("mpris", "setLoop", function (message) {
     }
 });
 
+addCallback("mpris", "identify", function (message) {
+    if (activePlayer) {
+        // We don't have a dedicated "send player info" callback, so we instead send a "playing"
+        // and if we're paused, we'll send a "paused" event right after
+        // TODO figure out a way how to add this to the host without breaking compat
+
+        var paused = activePlayer.paused;
+        playerPlaying(activePlayer);
+        if (paused) {
+            playerPaused(activePlayer);
+        }
+    }
+});
+
 function playerPlaying(player) {
     setPlayerActive(player);
 }
@@ -213,6 +241,13 @@ function setPlayerActive(player) {
     });
 }
 
+function sendPlayerGone() {
+    activePlayer = undefined;
+    playerMetadata = {};
+    playerCallbacks = [];
+    sendMessage("mpris", "gone");
+}
+
 function sendPlayerInfo(player, event, payload) {
     if (player != activePlayer) {
         return;
@@ -244,6 +279,13 @@ function registerPlayer(player) {
         sendPlayerInfo(player, "waiting");
     });
 
+    // playlist is now empty or being reloaded, stop player
+    // e.g. when using Ajax page navigation and the user nagivated away
+    player.addEventListener("emptied", function () {
+        // could have its own signal but for compat it's easier just to pretend to have stopped
+        sendPlayerInfo(player, "stopped");
+    });
+
     // opposite of "waiting", we finished buffering enough
     // only if we are playing, though, should we set playback state back to playing
     player.addEventListener("canplay", function () {
@@ -340,10 +382,7 @@ function playerPause() {
     }
 }
 
-document.addEventListener("DOMContentLoaded", function() {
-
-    registerAllPlayers();
-
+function loadMpris() {
     // TODO figure out somehow when a <video> tag is added dynamically and autoplays
     // as can happen on Ajax-heavy pages like YouTube
     // could also be done if we just look for the "audio playing in this tab" and only then check for player?
@@ -368,6 +407,27 @@ document.addEventListener("DOMContentLoaded", function() {
                     registerPlayer(player);
                 });
             });
+
+            mutation.removedNodes.forEach(function (node) {
+                if (typeof node.matches !== "function" || typeof node.querySelectorAll !== "function") {
+                    return;
+                }
+
+                if (node.matches("video,audio")) {
+                    if (node == activePlayer) {
+                        sendPlayerGone();
+                    }
+                    return;
+                }
+
+                var players = node.querySelectorAll("video,audio");
+                players.forEach(function (player) {
+                    if (player == activePlayer) {
+                        sendPlayerGone();
+                        return;
+                    }
+                });
+            });
         });
     });
 
@@ -401,14 +461,17 @@ document.addEventListener("DOMContentLoaded", function() {
         // about to navigate to a different page, tell our extension that the player will be gone shortly
         // we listen for tab closed in the extension but we don't for navigating away as URL change doesn't
         // neccesarily mean a navigation but beforeunload *should* be the thing we want
-
-        activePlayer = undefined;
-        playerMetadata = {};
-        playerCallbacks = [];
-        sendMessage("mpris", "gone");
+        sendPlayerGone();
     });
 
-});
+    // In some cases DOMContentLoaded won't fire, e.g. when watching a video file directly in the browser
+    // it generates a "video player" page for you but won't fire the event
+    registerAllPlayers();
+
+    document.addEventListener("DOMContentLoaded", function() {
+        registerAllPlayers();
+    });
+}
 
 // This adds a shim for the Chrome media sessions API which is currently only supported on Android
 // Documentation: https://developers.google.com/web/updates/2017/02/media-session
@@ -423,111 +486,126 @@ document.addEventListener("DOMContentLoaded", function() {
 // the rest is only set up in DOMContentLoaded which is only executed for proper pages anyway
 
 // tagName always returned "HTML" for me but I wouldn't trust it always being uppercase
-if (document.documentElement.tagName.toLowerCase() === "html") {
-    executeScript(`
-        function() {
-            ${mediaSessionsClassName} = function() {};
-            ${mediaSessionsClassName}.callbacks = {};
-            ${mediaSessionsClassName}.metadata = {};
-            ${mediaSessionsClassName}.playbackState = "none";
-            ${mediaSessionsClassName}.sendMessage = function(action, payload) {
-                var transferItem = document.getElementById('${mediaSessionsTransferDivId}');
-                transferItem.innerText = JSON.stringify({action: action, payload: payload});
+function loadMediaSessionsShim() {
+    if (document.documentElement.tagName.toLowerCase() === "html") {
+        executeScript(`
+            function() {
+                ${mediaSessionsClassName} = function() {};
+                ${mediaSessionsClassName}.transferItem = null;
+                ${mediaSessionsClassName}.callbacks = {};
+                ${mediaSessionsClassName}.metadata = {};
+                ${mediaSessionsClassName}.playbackState = "none";
+                ${mediaSessionsClassName}.sendMessage = function(action, payload) {
+                    this.transferItem.innerText = JSON.stringify({action: action, payload: payload});
 
-                var event = document.createEvent('CustomEvent');
-                event.initEvent('payloadChanged', true, true);
-                transferItem.dispatchEvent(event);
-            };
-            ${mediaSessionsClassName}.executeCallback = function (action) {
-                this.callbacks[action]();
-            };
+                    var event = document.createEvent('CustomEvent');
+                    event.initEvent('payloadChanged', true, true);
+                    this.transferItem.dispatchEvent(event);
+                };
+                ${mediaSessionsClassName}.executeCallback = function (action) {
+                    this.callbacks[action]();
+                };
 
-            navigator.mediaSession = {};
-            navigator.mediaSession.setActionHandler = function (name, cb) {
-                if (cb) {
-                    ${mediaSessionsClassName}.callbacks[name] = cb;
-                } else {
-                    delete ${mediaSessionsClassName}.callbacks[name];
-                }
-                ${mediaSessionsClassName}.sendMessage("callbacks", Object.keys(${mediaSessionsClassName}.callbacks));
-            };
-            Object.defineProperty(navigator.mediaSession, "metadata", {
-                get: function() { return ${mediaSessionsClassName}.metadata; },
-                set: function(newValue) {
-                    ${mediaSessionsClassName}.metadata = newValue;
-                    ${mediaSessionsClassName}.sendMessage("metadata", newValue.data);
-                }
-            });
-            Object.defineProperty(navigator.mediaSession, "playbackState", {
-                get: function() { return ${mediaSessionsClassName}.playbackState; },
-                set: function(newValue) {
-                    ${mediaSessionsClassName}.playbackState = newValue;
-                    ${mediaSessionsClassName}.sendMessage("playbackState", newValue);
-                }
-            });
+                navigator.mediaSession = {};
+                navigator.mediaSession.setActionHandler = function (name, cb) {
+                    if (cb) {
+                        ${mediaSessionsClassName}.callbacks[name] = cb;
+                    } else {
+                        delete ${mediaSessionsClassName}.callbacks[name];
+                    }
+                    ${mediaSessionsClassName}.sendMessage("callbacks", Object.keys(${mediaSessionsClassName}.callbacks));
+                };
+                Object.defineProperty(navigator.mediaSession, "metadata", {
+                    get: function() { return ${mediaSessionsClassName}.metadata; },
+                    set: function(newValue) {
+                        ${mediaSessionsClassName}.metadata = newValue;
+                        ${mediaSessionsClassName}.sendMessage("metadata", newValue.data);
+                    }
+                });
+                Object.defineProperty(navigator.mediaSession, "playbackState", {
+                    get: function() { return ${mediaSessionsClassName}.playbackState; },
+                    set: function(newValue) {
+                        ${mediaSessionsClassName}.playbackState = newValue;
+                        ${mediaSessionsClassName}.sendMessage("playbackState", newValue);
+                    }
+                });
 
-            window.MediaMetadata = function (data) {
-                this.data = data;
-            };
-        }
-    `);
-
-    // here we replace the document.createElement function with our own so we can detect
-    // when an <audio> tag is created that is not added to the DOM which most pages do
-    // while a <video> tag typically ends up being displayed to the user, audio is not.
-    // HACK We cannot really pass variables from the page's scope to our content-script's scope
-    // so we just blatantly insert the <audio> tag in the DOM and pick it up through our regular
-    // mechanism. Let's see how this goes :D
-
-    executeScript(`function() {
-            var oldCreateElement = document.createElement;
-            document.createElement = function () {
-                var createdTag = oldCreateElement.apply(this, arguments);
-
-                var tagName = arguments[0];
-
-                if (typeof tagName === "string" && tagName.toLowerCase() === "audio") {
-                    (document.head || document.documentElement).appendChild(createdTag);
-                }
-
-                return createdTag;
-            };
-        }
-    `);
-
-    // now the fun part of getting the stuff from our page back into our extension...
-    // cannot access extensions from innocent page JS for security
-    var transferItem = document.createElement("div");
-    transferItem.setAttribute("id", mediaSessionsTransferDivId);
-    transferItem.style.display = "none";
-
-    (document.head || document.documentElement).appendChild(transferItem);
-
-    transferItem.addEventListener('payloadChanged', function() {
-        var json = JSON.parse(this.innerText);
-
-        var action = json.action
-
-        if (action === "metadata") {
-            // FIXME filter metadata, this stuff comes from a hostile environment after all
-
-            playerMetadata = json.payload;
-
-            sendMessage("mpris", "metadata", json.payload);
-        } else if (action === "playbackState") {
-            var playbackState = json.payload;
-
-            if (activePlayer) {
-                if (playbackState === "playing") {
-                    playerPlaying(activePlayer);
-                } else if (playbackState === "paused") {
-                    playerPaused(activePlayer);
-                }
+                window.MediaMetadata = function (data) {
+                    this.data = data;
+                };
             }
+        `);
 
-        } else if (action === "callbacks") {
-            playerCallbacks = json.payload;
-            sendMessage("mpris", "callbacks", json.payload);
-        }
-    });
+        // here we replace the document.createElement function with our own so we can detect
+        // when an <audio> tag is created that is not added to the DOM which most pages do
+        // while a <video> tag typically ends up being displayed to the user, audio is not.
+        // HACK We cannot really pass variables from the page's scope to our content-script's scope
+        // so we just blatantly insert the <audio> tag in the DOM and pick it up through our regular
+        // mechanism. Let's see how this goes :D
+
+        executeScript(`function() {
+                var oldCreateElement = document.createElement;
+                document.createElement = function () {
+                    var createdTag = oldCreateElement.apply(this, arguments);
+
+                    var tagName = arguments[0];
+
+                    if (typeof tagName === "string") {
+                        if (tagName.toLowerCase() === "audio" || tagName.toLowerCase() === "video") {
+                            (document.head || document.documentElement).appendChild(createdTag);
+                            createdTag.parentNode.removeChild(createdTag);
+                        }
+                    }
+
+                    return createdTag;
+                };
+            }
+        `);
+
+        // now the fun part of getting the stuff from our page back into our extension...
+        // cannot access extensions from innocent page JS for security
+        var transferItem = document.createElement("div");
+        transferItem.setAttribute("id", mediaSessionsTransferDivId);
+
+        // Add the element to the DOM so the page sees it
+        (document.head || document.documentElement).appendChild(transferItem);
+
+        // now keep a reference to it inside the page
+        executeScript(`
+            function() {
+                ${mediaSessionsClassName}.transferItem = document.getElementById('${mediaSessionsTransferDivId}');
+            }
+        `);
+
+        // and remove it again so the website isn't confused
+        transferItem.parentNode.removeChild(transferItem);
+
+        transferItem.addEventListener('payloadChanged', function() {
+            var json = JSON.parse(this.innerText);
+
+            var action = json.action
+
+            if (action === "metadata") {
+                // FIXME filter metadata, this stuff comes from a hostile environment after all
+
+                playerMetadata = json.payload;
+
+                sendMessage("mpris", "metadata", json.payload);
+            } else if (action === "playbackState") {
+                var playbackState = json.payload;
+
+                if (activePlayer) {
+                    if (playbackState === "playing") {
+                        playerPlaying(activePlayer);
+                    } else if (playbackState === "paused") {
+                        playerPaused(activePlayer);
+                    }
+                }
+
+            } else if (action === "callbacks") {
+                playerCallbacks = json.payload;
+                sendMessage("mpris", "callbacks", json.payload);
+            }
+        });
+    }
 }
